% -*- mode: slang_mode; mode: fold; -*-
% sltabc.sl ks 2002-05-30, 2002-06-23
% Notes                                                                %{{{
%_debug_info=1;
%_traceback=1;
%_slangtrace=1;
#iffalse
% Example to implement in jed.rc or .jedrc

% Adapt SLTABC to your environment.
% SLTABC should point to a directory,
% where this file and minihelp.dat is located.
variable SLTABC= path_concat( Jed_Home_Directory, "patchlib");

autoload( "sltabc", path_concat( SLTABC, "sltabc.sl"));
autoload( "sltabc_evaluate_cmd", path_concat( SLTABC, "sltabc.sl"));

% Adapt following setkey.
% This will invoke SLang prompt with tab completion.
"sltabc_evaluate_cmd"; ""; setkey();


% To use tab completion in SLang buffers
% a set_buffer_hook should be defined as follows.
define global_mode_hook (hook_name)
{
   if ( hook_name == "slang_mode_hook")
     {
        set_buffer_hook( "indent_hook", "sltabc");
     }
}
#endif
%}}}
require( "cmode.sl");
static variable Popup= "*sl-Completions*";
static variable Color_Lines;
static variable Color_Objs= [ % line and column colors                 %{{{
                             color_number( "comment"),
                             color_number( "menu_selection"),
                             color_number( "message"),
                             color_number( "error")];
%}}}
static variable intf=[  % 65 internal functions                        %{{{
                      "backward_delete_char",
                      "backward_delete_char_untabify",
                      "beep",
                      "beg_of_buffer",
                      "beg_of_line",
                      "begin_macro",
                      "center_line",
                      "copy_region",
                      "delete_char_cmd",
                      "delete_window",
                      "digit_arg",
                      "end_macro",
                      "end_of_buffer",
                      "enlarge_window",
                      "eol_cmd",
                      "evaluate_cmd",
                      "exchange",
                      "execute_macro",
                      "exit_jed",
                      "exit_mini",
                      "find_file",
                      "format_paragraph",
                      "goto_match",
                      "indent_line",
                      "insert_file",
                      "kbd_quit",
                      "kill_buffer",
                      "kill_line",
                      "kill_region",
                      "macro_query",
                      "mark_spot",
                      "mini_complete",
                      "mouse_cmd",
                      "narrow",
                      "narrow_paragraph",
                      "narrow_to_region",
                      "newline",
                      "newline_and_indent",
                      "next_char_cmd",
                      "next_line_cmd",
                      "one_window",
                      "other_window",
                      "page_down",
                      "page_up",
                      "pop_spot",
                      "previous_char_cmd",
                      "previous_line_cmd",
                      "quoted_insert",
                      "redraw",
                      "save_some_buffers",
                      "scroll_left",
                      "scroll_right",
                      "select_menubar",
                      "self_insert_cmd",
                      "set_mark_cmd",
                      "split_window",
                      "switch_to_buffer",
                      "sys_spawn_cmd",
                      "text_smart_quote",
                      "trim_whitespace",
                      "undo",
                      "widen",
                      "widen_region",
                      "write_buffer",
                      "yank"];
%}}}   
static variable keyw=[  % 49 reserved keywords                         %{{{
                      "ERROR_BLOCK",         % !if not used here
                      "EXECUTE_ERROR_BLOCK", % from syntax table
                      "EXIT_BLOCK",
                      "NULL",                % from syntax table
                      "USER_BLOCK0",
                      "USER_BLOCK1",
                      "USER_BLOCK2",
                      "USER_BLOCK3",
                      "USER_BLOCK4",
                      "__tmp",
                      "_for",
                      "abs",
                      "and",
                      "andelse",
                      "break",
                      "case",
                      "chs",
                      "continue",
                      "define",
                      "do",
                      "do_while",
                      "else",
                      "exch",
                      "for",
                      "foreach",
                      "forever",
                      "if",
                      "loop",
                      "mod",
                      "mul2",
                      "not",
                      "or",
                      "orelse",
                      "pop",
                      "private",
                      "public",
                      "return",
                      "shl",
                      "shr",
                      "sign",
                      "sqr",
                      "static",
                      "struct",
                      "switch",
                      "typedef",
                      "using",
                      "variable",
                      "while",                      
                      "xor"];
%}}}

static define subarray_by_re( a, r) % get subarray by regexp           %{{{
{
   variable x=Integer_Type[length(a)], i;
   
   for (i=0;i<length( a);i++)
     x[i]= string_match( a[i], r, 1);

   return a[where (x)];
}
%}}}
static define possible_words( w) % returns array of possible words     %{{{
{
   variable n, n1, n2, n3, a1, a2, a3;

   a1= _apropos( "Global", "^"+w, 0xF);
   a2= subarray_by_re( intf, "^"+w);
   a3= subarray_by_re( keyw, "^"+w);

   n1= length( a1);
   n2= length( a2);
   n3= length( a3);
   
   variable a= String_Type[n1+n2+n3];
   if ( n1) a[[:n1-1]]= a1;
   if ( n2) a[[n1:n1+n2-1]]= a2;
   if ( n3) a[[n1+n2:]]= a3;

   return a[array_sort(a)];;
}
%}}}

static define goto_popup() %                                           %{{{
{
%  provides a visible buffer %{{{
%  does not change current window
%  saves data to restore covered
%  or splitted window in blocal variables
% Now we are in the (normal) current window
% If we have 2 or more normal windows, we go to the next.
% (Thus the current remains visible.)

% popup if we have just 1 normal window

% NB: If mini is active, it increases nwindows by one.
%     Mini inserts itself BEFORE the current window.
%     Native window order is from top to bottom.
%}}}

   variable b, na, nn, v, nr, w_b, w_i;

   b= whatbuf();
   na= nwindows();                     % number of active windows
   nn= na -MINIBUFFER_ACTIVE; % number of normal windows

   v= buffer_visible( Popup);
   nr= na;
   loop ( na)
     {
        otherwindow();
        nr--;
        if ( whatbuf() == Popup) return nr;
        if ( v) continue;
        !if (andelse{ MINIBUFFER_ACTIVE}{ whatbuf() == " <mini>"}) break;
     }

   w_b= whatbuf();
   w_i= [ 0, what_line(), _get_point(), window_line()];
   if ( nn == 1)
     {
        w_i[0]= 1;
        splitwindow();
        otherwindow();
        nr= 1;
        if ( MINIBUFFER_ACTIVE)
          {
             if ( whatbuf() == " <mini>")
               {
                  otherwindow();
                  if ( b == " <mini>") nr= 2;
               }
             else
               {
                  if ( b != " <mini>") nr= 2;
               }
          }
     }

   sw2buf( Popup);
   create_blocal_var( "w_b");
   create_blocal_var( "w_i");
   set_blocal_var( w_b, "w_b");
   set_blocal_var( w_i, "w_i");
   return nr;
}
%}}}
static define remove_popup() %                                         %{{{
{
   variable b, na, nn, w_b, w_i;
   
   b= whatbuf();
   na= nwindows();                     % number of active windows
   nn= na -MINIBUFFER_ACTIVE; % number of normal windows

   loop ( na)
     {
        otherwindow();
        if ( whatbuf() != Popup) continue;
        !if (andelse{ blocal_var_exists( "w_b")}
              {blocal_var_exists( "w_i")}) continue;
        w_b= get_blocal_var( "w_b");
        w_i= get_blocal_var( "w_i");

        if (andelse{ w_i[0]}{ nn == 2})
          {
             otherwindow();
             if (andelse{ MINIBUFFER_ACTIVE}{ whatbuf() == " <mini>"})
               otherwindow();
             onewindow();
             if (andelse{ whatbuf() == w_b}{ what_line() == w_i[1]})
               recenter( w_i[3]);
             if ( b == " <mini>") otherwindow();
             break;
          }

        sw2buf( w_b);
        goto_line( w_i[1]);
        _set_point( w_i[2]);
        recenter( w_i[3]);
     }
   if ( bufferp( Popup)) delbuf( Popup);
}
%}}}

static define expand_word( w, a) %                                     %{{{
% General function for tab completion and similar.
% w: word to complete (string)
% a: possible words (string array)
% x: expansion of word (return value, string)

% NB: all possible words have to start with w
{
   variable x, m, i, k, n;

   if (not length(a)) return "";

%-- minimum strlen of all a[i]'s
   m= strlen(a[0]);
   for ( i= 1; i < length( a); i++)
     if ( n= strlen( a[i]), n < m) m= n;

%-- NB: w+x is the greatest common initial
%-- substring of all a[i]'s
   x= "";
   for ( n= strlen( w); n < m; n++)
     {
        k= a[0][[n:n]];
        for ( i =1; i < length(a); i++)
          if ( a[i][[n:n]] != k) return x;
        x+= k;
     }
   return x;
}
%}}}
static define tab_action( r) %                                         %{{{
{
   variable i, k, cw= "a-zA-Z0-9_";
     
   i= _get_point();
   if ( what_keymap() != "Mini_Map")
     {
        %-- do c_indent_line() and see what happens
        k= what_char();
        c_indent_line();
        
        if (orelse{ _get_point() != i}{ what_char() != k}) return 1;
     }

   bskip_chars( cw); push_mark(); skip_chars( cw); @r= bufsubstr();

   if ( @r != "") return 0;
   if (andelse{ what_keymap() == "Mini_Map"}{ eolp()}) return 0;
   skip_chars( "^"+cw);
   return 1;
}
%}}}

static define minihelp_text( w, m, r) %                                %{{{
{
   variable t, f;
   
   if ( m) return 0;
   f= path_concat( SLTABC, "minihelp.dat");
   t= get_doc_string_from_file( f, w);
   if ( t == NULL) return 0;

   t= strtrim( t);  % remove newline
   (t,)= strreplace( t, w, "", 1);
   (t,)= strreplace( t, "~", w, 1);
   @r= substr( t, 2, -1);
   return 1;
}
%}}}
static define auto_comment( w, i) %                                    %{{{
{
   variable m;
   variable t= ["Library/User Variable",
                "Intrinsic/Jed Variable",
                "",
                "Intrinsic/Jed Function",
                "Library/User Function"];

   if (not i) if ( m= is_defined( w), m) return t[m+2];
   if ( is_internal( w)) return "Internal Function";
   if (not i) return "Reserved Keyword";
   return "Object type not detectable";
}
%}}}
static define insert_matches( w, a) %                                  %{{{
{
   variable k, i, t, c= Mark_Type[99];

   k= 0;
   bob(); insert( ">>>");
   for (i=0;i<length(a);i++)
     {
        goto_column( 7);
        if ( a[i] == w)
          {
             c[k]= create_line_mark( Color_Objs[0]); k++;
             if ( minihelp_text( w, i, &t))
               {
                  insert( t+"\n");
                  goto_column( 9 +strlen(w));
                  c[k]= create_line_mark( Color_Objs[0]); k++;
                  insert( auto_comment( w, i)+"\n");
               }
             else
               insert( w +"  "+auto_comment( w, i)+"\n");
          }
        else
          insert( a[i]+"\n");
     }

   ()= left(1); del();
   bol(); deln( 3); insert( ">>>");

   if ( length(a))
     set_column_colors( Color_Objs[2], 1, 80);
   else
     {
        set_column_colors( Color_Objs[3], 1, 80);
        goto_column( 7);
        insert( w +"  --  no match found"); w= "";
     }
   set_column_colors( Color_Objs[1], 1, 6+strlen(w));

   Color_Lines= c;
}
%}}}
static define show_matches( w, a) %                                    %{{{
{
   variable nr, i;
   
   nr= goto_popup();
   erase_buffer();

   insert_matches( w, a);
   bob();

   set_buffer_modified_flag( 0);
   set_status_line( "(Jed %v) PageDown/PageUp/Tab: %b %l/%L",0);

   loop ( nr) otherwindow();
}
%}}}

% slang prompt
static variable e_f;
static define set_keys_mini( j) %                                      %{{{
{
   variable f, s;
   variable mini = "Mini_Map";

   if ( j)
     {
        definekey( "sltabc", "\t", mini);
        definekey( "sltabc_exit_minibuffer", "\r",  mini);
     }
   else
     {
        definekey( "mini_complete", "\t", mini);
        definekey( "mini_exit_minibuffer", "\r", mini);
     }

   foreach ([ "page_down", "page_up"])
     {
        f= ();
        s= "sltabc_" +f;
        if ( j)
          loop ( which_key( f)) { s; exch(); mini; definekey();}
        else
          loop ( which_key( s)) { f; exch(); mini; definekey();}
     }

}
%}}}
public define sltabc_exit_minibuffer() %                               %{{{
{
   variable b;

   eob();
   e_f= 0;
   while ( andelse{ left(1)}{ what_char() == '='}{ e_f < 2})
     {
        e_f++;
        del();
     }

   set_keys_mini( 0);
   remove_popup();
   mini_exit_minibuffer();
}
%}}}
public define sltabc_evaluate_cmd() %                                  %{{{
{
   variable s;
   
   set_keys_mini( 1);
   ERROR_BLOCK
     {
        _clear_error();
        set_keys_mini( 0);
        remove_popup();
        return;
     }
   call( "evaluate_cmd");
   
   switch ( e_f)
     {  case 1: string(); message();}
     {  case 2:
        _stk_reverse( _stkdepth);
        s= "";
        while ( _stkdepth) s+= string()+";";
        message( s);
     }
}
%}}}

% slang buffer and slang prompt
static define set_keys( j) %                                           %{{{
{
   variable f, s;

   if ( what_keymap() == "Mini_Map") return;
   foreach ([ "page_down", "page_up"])
     {
        f= ();
        s= "sltabc_" +f;
        if ( j)
          loop ( which_key( f)) { s; exch(); local_setkey();}
        else
          loop ( which_key( s)) { f; exch(); local_setkey();}
     }
}
%}}} insert
public define sltabc() %                                               %{{{
{
   variable w, x, a, h;

   if ( tab_action( &w))
     {
        if ( buffer_visible( Popup))
          {
             remove_popup();
             set_keys( 0);
          }
        return;
     }
   
   a= possible_words( w);
   x= expand_word( w, a);

   if ( what_keymap() != "Mini_Map")
     h= "\t";
   else
     h= _function_name();

   if (andelse{ x == ""}
         { LAST_KBD_COMMAND == h}
         { buffer_visible( Popup)})
     {
        remove_popup();
        set_keys( 0);
     }
   else
     {
        insert( x);
        w+= x;
        show_matches( w, a);
        set_keys( 1);
     }
}
%}}}
static define sltabc_page1( f, nr) %{{{
{
   variable j, k;

   k= ( what_keymap() != "Mini_Map");
   if ( j= not buffer_visible( Popup), j)
     {
        set_keys( 0);
        call( f);
     }
   else
     @nr= goto_popup();
   return j;
}
%}}}
static define sltabc_page2( j, nr) %{{{
{
   loop ( nr) otherwindow();
   if (not j) return;
   remove_popup();
   set_keys( 0);
}
%}}}
public define sltabc_page_down() %                                     %{{{
{
   variable nr, j, n, f= "page_down";

   if ( sltabc_page1( f, &nr)) return;
   n= window_info( 'r');
   ()= down( n -window_line()); eol();   
   if ( j= eobp(), not j)
     {
        call( f);
        ()= down( n -window_line());
        eol();
        recenter( n);
     }
   sltabc_page2( j, nr);
}
%}}}
public define sltabc_page_up() %                                       %{{{
{
   variable nr, j, f= "page_up";

   if ( sltabc_page1( f, &nr)) return;
   if ( j= ( what_line() == window_line()), not j)
     {
        call( f);
        ()= up( window_line() -1);
        recenter( 1);
     }
   sltabc_page2( j, nr);
}
                                 
%}}}
